# SPDX-FileCopyrightText: Copyright (c) 2023-2024, Kr8s Developers (See LICENSE for list)
# SPDX-License-Identifier: BSD 3-Clause License
from kr8s._api import Api  # noqa

from ._api import api  # noqa
from ._helpers import api_resources, get, version, watch, whoami  # noqa
